use rppal::{i2c, i2c::I2c};
use std::io::{self, stdin, stdout, BufRead, Write};

const C0: u8 = 7;
const A0: u8 = 6;

#[allow(unused)]
#[repr(u8)]
enum LcdCommand {
    Clear = 1,
    ReturnHome = 2,
    DisplayOn = 1 << 3 | 1 << 2 | 1 << 1 | 1 << 0,
    FunctionSet = 1 << 5 | 1 << 4 | 1 << 3,
    EntryModeSet = 1 << 2 | 1 << 1,
}

#[allow(unused)]
#[repr(u8)]
enum ControlByte {
    Normal = 1 << C0,
    Last = 0 << C0,
    Data = 0 << C0 | 1 << A0,
}

struct Lcd {
    bus: I2c,
}

impl Lcd {
    pub fn new() -> Result<Self, i2c::Error> {
        let mut bus = I2c::with_bus(1)?;
        bus.set_slave_address(60)?;
        let setup = [
            ControlByte::Last as u8,
            LcdCommand::FunctionSet as u8,
            LcdCommand::DisplayOn as u8,
            LcdCommand::EntryModeSet as u8,
            LcdCommand::Clear as u8,
        ];
        bus.write(&setup)?;
        Ok(Self { bus })
    }

    pub fn write(&mut self, value: &str) -> Result<(), i2c::Error> {
        let mut data = vec![ControlByte::Data as u8];
        data.extend(value.bytes());
        let _ = self.bus.write(&data)?;
        Ok(())
    }

    pub fn set_cursor(&mut self, col: u8, row: u8) -> Result<(), i2c::Error> {
        let pos = col | (row * 0x40) | 1 << 7;
        let data = vec![ControlByte::Last as u8, pos];
        let _ = self.bus.write(&data)?;
        Ok(())
    }

    pub fn clear(&mut self) -> Result<(), i2c::Error> {
        let data = vec![ControlByte::Last as u8, LcdCommand::Clear as u8];
        let _ = self.bus.write(&data)?;
        Ok(())
    }
}

fn main() -> io::Result<()> {
    let mut lcd = Lcd::new().unwrap();
    loop {
        print!("> ");
        stdout().lock().flush()?;
        let mut line = String::new();
        stdin().lock().read_line(&mut line)?;
        if line.starts_with("clear") {
            lcd.clear().unwrap();
            lcd.set_cursor(0, 0).unwrap();
        } else if line.starts_with("print") {
            lcd.write(&line[6..]).unwrap();
        } else {
            println!("unrecognized command \"{line}\"");
        }
    }
}
